/*  Pcsx - Pc Psx Emulator
 *  Copyright (C) 1999-2002  Pcsx Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <dlfcn.h>
#include <sys/mman.h>
#include <errno.h>
#include <string.h>
#include <sys/time.h>
#include <pthread.h>
#include <dirent.h>

#include <SDL/SDL.h>

#include "Linux.h"
#include "Sio.h"

#include "minimal.h"

int UseGui = 0;
int CancelQuit = 0;
SDL_Surface* sdlscreen;
char* packfile = NULL;
unsigned long mStopEmulation = 0;
char gamepath[256];
long LoadCdBios;
PcsxConfig Config;
FILE* emuLog;
int Log=0;
unsigned long gpuDisp;
int cdOpenCase = 0;
int ShowPic=0;
int StatesC = 0;

#if defined(__BEOS__) || defined(__LINUX__)
static void ChangeWorkingDirectory(char *exe)
{
	char *s = strrchr(exe, '/');
	if (s != NULL) {
		*s = '\0';
		chdir(exe);
		*s = '/';
	}
}
#endif

struct dir_item {
	char *name;
	int type; // 0=dir, 1=file, 2=zip archive
};

void sort_dir(struct dir_item *list, int num_items, int sepdir) {
	int i;
	struct dir_item temp;

	for(i=0; i<(num_items-1); i++) {
		if(strcmp(list[i].name, list[i+1].name)>0) {
			temp=list[i];
			list[i]=list[i+1];
			list[i+1]=temp;
			i=0;
		}
	}
	if(sepdir) {
		for(i=0; i<(num_items-1); i++) {
			if((list[i].type!=0)&&(list[i+1].type==0)) {
				temp=list[i];
				list[i]=list[i+1];
				list[i+1]=temp;
				i=0;
			}
		}
	}
}

char *FileReq(char *dir)
{
	static char *cwd=NULL;
	static int cursor_pos=1;
	static int first_visible;
	static int num_items=0;
	DIR *dirstream;
	struct dirent *direntry;
	struct dir_item dir_items[1024];
	char *path;
	struct stat item;
	static int row;
	int pathlength;
	char tmp_string[32];
	char *selected;
	unsigned char* keys;
	//int numkeys;
#define MENU_Y 25
#define MENU_LS MENU_Y+10
#define MENU_HEIGHT 15
	if(dir!=NULL) cwd=dir;
	if(cwd==NULL) cwd=gamepath;
	
	gp2x_printf(NULL, 0, 0, "ZPSX v0.06 BY ZODTTD");

	if(num_items==0) {
			int i2;
			dirstream=opendir(cwd);
			if(dirstream==NULL) {
				return NULL;
			}
			// read directory entries
			while(direntry=readdir(dirstream)) {
				dir_items[num_items].name=(char *)malloc(strlen(direntry->d_name)+1);
				strcpy(dir_items[num_items].name, direntry->d_name);
				num_items++;
				if(num_items>1024) break;
			}
			closedir(dirstream);
			// get entry types
			for(i2=0; i2<num_items; i2++) {
				path=(char *)malloc(strlen(cwd)+strlen(dir_items[i2].name)+2);
				sprintf(path, "%s/%s", cwd, dir_items[i2].name);
				if(!stat(path, &item)) {
					if(S_ISDIR(item.st_mode)) {
						dir_items[i2].type=0;
					} 
					else 
					{
						int len = strlen(dir_items[i2].name);
						
						dir_items[i2].type = 2;
						
						if (len >= 4)
						{
							if (!strnicmp(dir_items[i2].name+(len-2), ".Z", 2)) 
							{
								dir_items[i2].type = 1;
							}
							if (!strnicmp(dir_items[i2].name+(len-4), ".bin", 4)) 
							{
								dir_items[i2].type = 1;
							}
							if (!strnicmp(dir_items[i2].name+(len-4), ".ZNX", 4)) 
							{
								dir_items[i2].type = 1;
							}
						}
					}
				} else {
					dir_items[i2].type=0;
				}
				free(path);
			}
		sort_dir(dir_items, num_items, 1);
		cursor_pos=0;
		first_visible=0;
	}

	// display current directory
    //gp2x_printf(NULL, 80, MENU_Y, blank_line);
    gp2x_printf(NULL, 80, MENU_Y, cwd);

	// display directory contents
	row=0;
	while(row<num_items && row<MENU_HEIGHT) {
		if(row==(cursor_pos-first_visible)) {
			// draw cursor
			gp2x_printf(NULL, 80, MENU_LS+(10*row), "------>");

			selected=dir_items[row+first_visible].name;
		}
		//gp2x_printf(NULL, 80, MENU_LS+(10*row), blank_line);
		if(dir_items[row+first_visible].type==0) 
		{
        		gp2x_printf(NULL, 80, MENU_LS+(10*row), "DIR ");
		}
		snprintf(tmp_string, 30, "%s", dir_items[row+first_visible].name);
        	gp2x_printf(NULL, 80+(10*6), MENU_LS+(10*row), tmp_string);
		row++;
	}
	while(row<MENU_HEIGHT) {
        //gp2x_printf(NULL, 80, MENU_LS+(10*row), blank_line);
		row++;
	}

	// pump events so GetKeyState works
	SDL_PumpEvents();

	keys = SDL_GetKeyState(NULL);
	
	if ( keys[SDLK_BACKSPACE] ) { // exit
		// close SDL
		SDL_Quit();

		exit(0);		
	}
	else if(keys[SDLK_DOWN]) { //down
			if(cursor_pos<(num_items-1)) cursor_pos++;
			if((cursor_pos-first_visible)>=MENU_HEIGHT) first_visible++;
	} 
	else if(keys[SDLK_UP]) { // up
			if(cursor_pos>0) cursor_pos--;
			if(cursor_pos<first_visible) first_visible--;
	} 
	else if(keys[SDLK_LEFT]) { //left
			if(cursor_pos>=10) cursor_pos-=10;
			else cursor_pos=0;
			if(cursor_pos<first_visible) first_visible=cursor_pos;
	} 
	else if(keys[SDLK_RIGHT]) { //right
			if(cursor_pos<(num_items-11)) cursor_pos+=10;
			else cursor_pos=num_items-1;
			if((cursor_pos-first_visible)>=MENU_HEIGHT) 
				first_visible=cursor_pos-(MENU_HEIGHT-1);
	} 
	else if(keys[SDLK_RETURN]) 
	{ // button 1
		int i;
		path=(char *)malloc(strlen(cwd)
			+strlen(dir_items[cursor_pos].name)
			+2);
		sprintf(path, "%s/%s", cwd, dir_items[cursor_pos].name);
		//FIXME ut oh
		//for(i=0; i<num_items; i++) free(dir_items[i].name);
		num_items=0;
		if(dir_items[cursor_pos].type==0) {
			// directory selected
			pathlength=strlen(path);
			if(path[pathlength-1]=='.'
				// check for . selected
					&& path[pathlength-2]=='/') {
				path[pathlength-2]='\0';
				cwd=path;
			} else if(path[pathlength-1]=='.'
				// check for .. selected
					&& path[pathlength-2]=='.'
					&& path[pathlength-3]=='/'
					&&pathlength>8) {
				for(i=4; path[pathlength-i]!='/'; i++);
				path[pathlength-i]='\0';
				if(strlen(path)<=5) {
					path[5]='/';
					path[6]='\0';
				}
				cwd=path;
			} else {
				cwd=path;
			}
			sleep(1);
		} else {
			// file selected
			return path;
		}
	}
	
	return NULL;
}

int main(int argc, char *argv[]) {
	char *file = NULL;
	char *lang;
	int runcd = 0;
	int loadst = 0;
	int i;
	unsigned char* keys;

#if defined(__BEOS__) || defined(__LINUX__)
	ChangeWorkingDirectory(argv[0]);
#endif
	getcwd(gamepath, 256);

	memset(&Config, 0, sizeof(PcsxConfig));
	strcpy(Config.Net,"Disabled");
	Config.PsxAuto = 1;
	Config.Cdda = 1;
	Config.Cpu = 1;
	Config.Mdec = 0;
	Config.PsxOut = 0;
	Config.PsxType = 0;
	Config.QKeys = 0;
	Config.RCntFix = 0;
	Config.Sio = 0;
	Config.SpuIrq = 0;
	Config.VSyncWA = 0;
	Config.Xa = 1;
	sprintf(Config.BiosDir, "%s", gamepath);
	sprintf(Config.Bios, "/scph1001.bin");
	sprintf(Config.Mcd1, "%s/mcd001.mcr", gamepath);
	sprintf(Config.Mcd2, "%s/mcd002.mcr", gamepath);

#if 0
	if (Config.Lang[0] == 0) {
		char *str = setlocale(LC_MESSAGES, NULL);
		char _lang[3];

		strncpy(_lang, str, 2); _lang[2] = 0;
		strcpy(Config.Lang, _lang);
	}

	langs = (_langs*)malloc(sizeof(_langs));
	strcpy(langs[0].lang, "en");
	InitLanguages(); i=1;
	while ((lang = GetLanguageNext()) != NULL) {
		langs = (_langs*)realloc(langs, sizeof(_langs)*(i+1));
		strcpy(langs[i].lang, lang);
		i++;
	}
	CloseLanguages();
	langsMax = i;

	if (UseGui) gtk_init(NULL, NULL);
#endif
	// init SDL
  	if(SDL_Init(SDL_INIT_VIDEO)) {
    		printf(0, "Failed to init SDL: %s", SDL_GetError());
    		exit(-1);
  	}
	SDL_ShowCursor(0);
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
	SDL_EnableUNICODE(1);
	
	// init display
	sdlscreen = SDL_SetVideoMode(DISPLAY_X, DISPLAY_Y, 16, SDL_HWSURFACE|SDL_HWPALETTE|SDL_FULLSCREEN);

	// init fonts 
	gp2x_init(1000, 16, 11025, 16, 1, 60, 1);

	// pick a game
	for(;;)
	{
		SDL_FillRect(sdlscreen, NULL, 0);
		packfile = FileReq(NULL);
		SDL_UpdateRect(sdlscreen, 0, 0, 0, 0);
		if( packfile != NULL )
		{
			break;
		}
		usleep(100000);
	}

	// clear screen
	SDL_FillRect(sdlscreen, NULL, 0);
	SDL_UpdateRect(sdlscreen, 0, 0, 0, 0);
	
	// pump events so GetKeyState works
	SDL_PumpEvents();

	keys = SDL_GetKeyState(NULL);
	if( keys[SDLK_h] )
	{
		LoadCdBios = 0;
		sprintf(Config.Bios, "HLE");
		// write text to both framebuffers
		gp2x_printf(NULL, 70, 100, "LOADING GAME USING HLE EMULATION");
		SDL_UpdateRect(sdlscreen, 0, 0, 0, 0);
	}
	else
	{
		LoadCdBios = 0;
		// write text to both framebuffers
		gp2x_printf(NULL, 70, 100, "LOADING GAME WITHOUT HLE EMULATION");
		gp2x_printf(NULL, 60, 110, "HOLD DOWN THE H KEY NEXT TIME FOR HLE");
		SDL_UpdateRect(sdlscreen, 0, 0, 0, 0);
	}

	if (SysInit() == -1) return 1;

#if 0 // UNIMPLEMENTED
	if (UseGui) {
		StartGui();
		return 0;
	}
#endif

	if (OpenPlugins() == -1) return 1;
	SysReset();
	CheckCdrom();
	LoadCdrom();

#if 0 // UNIMPLEMENTED
	if (file != NULL) Load(file);
	else {
		if (runcd == 1) {
			LoadCdBios = 0;
			if (LoadCdrom() == -1) {
				ClosePlugins();
				printf(_("Could not load Cdrom\n"));
				return -1;
			}
		} else if (runcd == 2) LoadCdBios = 1;
	}

	if (loadst) {
		char Text[256];
		StatesC = loadst-1;
		sprintf (Text, "sstates/%10.10s.%3.3d", CdromLabel, StatesC);
		LoadState(Text);
	}
#endif

	psxCpu->Execute();

	return 0;
}

#if 0 // UNIMPLEMENTED
DIR *dir;

void InitLanguages() {
	dir = opendir("Langs");
}

char *GetLanguageNext() {
	struct dirent *ent;

	if (dir == NULL) return NULL;
	for (;;) {
		ent = readdir(dir);
		if (ent == NULL) return NULL;

		if (!strcmp(ent->d_name, ".")) continue;
		if (!strcmp(ent->d_name, "..")) continue;
		break;
	}

	return ent->d_name;
}

void CloseLanguages() {
	if (dir) closedir(dir);
}

void ChangeLanguage(char *lang) {
	strcpy(Config.Lang, lang);
	SaveConfig();
	LoadConfig();
}
#endif

int SysInit() {

#ifdef GTE_DUMP
	gteLog = fopen("gteLog.txt","wb");
	setvbuf(gteLog, NULL, _IONBF, 0);
#endif

#ifdef EMU_LOG
#ifndef LOG_STDOUT
	emuLog = fopen("emuLog.txt","wb");
#else
	emuLog = stdout;
#endif
	setvbuf(emuLog, NULL, _IONBF, 0);
#endif

	psxInit();
	LoadPlugins();

	LoadMcds(Config.Mcd1, Config.Mcd2);

	return 0;
}

void SysReset() {
	psxReset();
}

void SysClose() {
	psxShutdown();
	ReleasePlugins();

	if (emuLog != NULL) fclose(emuLog);
}

void SysPrintf(char *fmt, ...) {
	va_list list;
	char msg[512];

	va_start(list, fmt);
	vsprintf(msg, fmt, list);
	va_end(list);

	if (Config.PsxOut) printf ("%s", msg);

	gp2x_printf(NULL, 10, 10, msg);
	SDL_UpdateRect(sdlscreen, 0, 0, 0, 0); 
#ifdef EMU_LOG
#ifndef LOG_STDOUT
	fprintf(emuLog, "%s", msg);
#endif
#endif
}

void *SysLoadLibrary(char *lib) {
	return dlopen(lib, RTLD_NOW);
}

void *SysLoadSym(void *lib, char *sym) {
	return dlsym(lib, sym);
}

const char *SysLibError() {
	return dlerror();
}

void SysCloseLibrary(void *lib) {
	dlclose(lib);
}

void SysUpdate() {
// EDIT	PADhandleKey(PAD1_keypressed());
// EDIT	PADhandleKey(PAD2_keypressed());
}

void SysRunGui() {
// EDIT	RunGui();
}

void SysMessage(char* fmt, ...) {
#ifdef EMU_LOG
#ifndef LOG_STDOUT
	va_list list;
	char msg[512];

	va_start(list, fmt);
	vsprintf(msg, fmt, list);
	va_end(list);

	if (Config.PsxOut) printf ("%s", msg);
	//SysPrintf("%s", msg);
	fprintf(emuLog, "%s\r\n", msg);
#endif
#endif
}
