#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "PsxCommon.h"
#include "plugins.h"
#include "PlugCD.h"

#include <zlib.h>

#define CD_FRAMESIZE_RAW	2352
#define DATA_SIZE	(CD_FRAMESIZE_RAW-12)

#define itob(i)		((i)/10*16 + (i)%10)	/* u_char to BCD */
#define btoi(b)		((b)/16*10 + (b)%16)	/* BCD to u_char */

#define MSF2SECT(m,s,f)	(((m)*60+(s)-2)*75+(f))

extern char* packfile; //[256];
int Zmode;
unsigned char *Ztable;
unsigned char Zbuf[CD_FRAMESIZE_RAW];

void UpdateZmode() {
	int len = strlen(CDConfiguration.fn);

	if (len >= 2) {
		if (!strnicmp(CDConfiguration.fn+(len-2), ".Z", 2)) {
			Zmode = 1; return;
		}
	}
	
	if (len >= 4) {
		if (!strnicmp(CDConfiguration.fn+(len-4), ".ZNX", 4)) {
			Zmode = 2; return;
		}
	}	
/*
	if (len >= 3) {
		if (!strnicmp(IsoFile+(len-3), ".bz", 3)) {
			Zmode = 2; return;
		}
	}
*/

	Zmode = 0;
}

// gets track 
long getTN(unsigned char* buffer)
{
 int numtracks = getNumTracks();


 if (-1 == numtracks)
 {
         return -1;
 }

 buffer[0]=1;
 buffer[1]=numtracks;

//   printf("getnumtracks %d %d\n", (int)buffer[0], (int)buffer[1]);
//SysPrintf("end getTn()\r\n");
   return 0;
}

 // if track==0 -> return total length of cd
 // otherwise return start in bcd time format
long getTD(int track, unsigned char* buffer)
{
      // lasttrack just keeps track of which track TD was requested last (go fig)
//	SysPrintf("start getTD()\r\n");

   if (track > CD.numtracks)
   {
//      printf("getTD bad %2d\n", track);
      return -1;
   }

   if (track == 0)
   {
      buffer[0] = CD.tl[track].end[0];
      buffer[1] = CD.tl[track].end[1];
      buffer[2] = CD.tl[track].end[2];
   }
   else
   {
      buffer[0] = CD.tl[track].start[0];
      buffer[1] = CD.tl[track].start[1];
      buffer[2] = CD.tl[track].start[2];
   }
//   printf("getTD %2d %02d:%02d:%02d\n", track, (int)buffer[0],
//         (int)buffer[1], (int)buffer[2]);

   // bcd encode it
   buffer[0] = itob(buffer[0]);
   buffer[1] = itob(buffer[1]);
   buffer[2] = itob(buffer[2]);
//   SysPrintf("end getTD()\r\n");
   return 0;
}

// opens a binary cd image and calculates its length
void openBin(const char* filename)
{
    long blocks;
	unsigned long size, end;
	
    CD.cd = fopen(filename, "rb");
	
    if (CD.cd == NULL)
    {
    	SysPrintf("Error Opening: %s", filename);
    }

    fseek(CD.cd, 0, SEEK_END);
    end = ftell(CD.cd);
    size = end;
	//VFSFileSize( CD.cd->ref, &size );
	
    rc = fseek(CD.cd, 0, SEEK_SET);
    blocks = size / 2352;

    // put the track length info in the track list
    CD.tl = (Track*) malloc(sizeof(Track));

    CD.tl[0].type = Mode2;
    CD.tl[0].num = 1;
    CD.tl[0].start[0] = 0;
    CD.tl[0].start[1] = 0;
    CD.tl[0].start[2] = 0;
    CD.tl[0].end[2] = blocks % 75;
    CD.tl[0].end[1] = ((blocks - CD.tl[0].end[2]) / 75) % 60;
    CD.tl[0].end[0] = (((blocks - CD.tl[0].end[2]) / 75) - CD.tl[0].end[1]) / 60;

    CD.tl[0].start[1] += 2;
    CD.tl[0].end[1] += 2;

    normalizeTime(CD.tl[0].end);

    CD.numtracks = 1;

//    SysPrintf("end openBin()\r\n");
}

void addBinTrackInfo()
{
//    SysPrintf("start addBinTrackInfo()\r\n");

    CD.tl = realloc(CD.tl, (CD.numtracks + 1) * sizeof(Track));
    CD.tl[CD.numtracks].end[0] = CD.tl[0].end[0];
    CD.tl[CD.numtracks].end[1] = CD.tl[0].end[1];
    CD.tl[CD.numtracks].end[2] = CD.tl[0].end[2];
    CD.tl[CD.numtracks].start[0] = CD.tl[0].start[0];
    CD.tl[CD.numtracks].start[1] = CD.tl[0].start[1];
    CD.tl[CD.numtracks].start[2] = CD.tl[0].start[2];

//    SysPrintf("end addBinTrackInfo()\r\n");
}

// new file types should be added here and in the CDOpen function
void newCD(const char * filename)
{
//    SysPrintf("start newCD()\r\n");

	if (Zmode) 
	{   
		FILE *f;
		char table[256];
		unsigned char* Ztemp;
		unsigned long size;
		unsigned long offset;
		
		CD.type = IndexZ;
		
		strcpy(table, filename);
		if (Zmode) strcat(table, ".table");
		else strcat(table, ".index");
/*
		if (stat(table, &buf) == -1) {		
			printf("Error loading %s\n", table);
			cdHandle = NULL;
			return 0;
		}
*/
		f = fopen(table, "rb");
		if( f == NULL )
		{
			SysPrintf("NewCD ERR 5 ");
			return;
		}
		fseek(f, 0, SEEK_END);
		size = ftell(f);
		fseek(f, 0, SEEK_SET);
		Ztable = (unsigned char*)malloc(size);
		if (Ztable == NULL) {
			SysPrintf("Needs %u more bytes of RAM.", size);
			return;
		}
		fread(Ztable, 1, size, f);
		fclose(f);		
					
    	CD.cd = fopen(filename, "rb");
	
 	   	if (CD.cd == NULL)
   	   	{
    		SysMessage("Error Opening : %s", filename);
       	}
     
     	CD.numtracks = 1;  	
	    CD.bufferPos = 0x7FFFFFFF;
	    seekSector(0,2,0);
	}
	else
	{    
	    CD.type = Bin;
	    openBin(filename);
	    addBinTrackInfo();

	    CD.bufferPos = 0x7FFFFFFF;
	    seekSector(0,2,0);
	}
//    SysPrintf("end newCD\r\n");
}


// return the sector address - the buffer address + 12 bytes for offset.
unsigned char* getSector()
{
//    SysPrintf("getSector()\r\n");
	
    return CD.buffer + (CD.sector - CD.bufferPos) + 12;
}

// returns the number of tracks
signed char getNumTracks()
{
//    SysPrintf("start getNumTracks()\r\n");
    // if there's no open cd, return -1
    if (CD.cd == 0) {
        return -1;
    }

//    printf("numtracks %d\n",CD.numtracks);
//    SysPrintf("end getNumTracks()\r\n");
    return CD.numtracks;
}

void readit(const unsigned char m, const unsigned char s, const unsigned char f)
{
	if (Zmode == 1) 
	{ //.Z
		unsigned long pos;
		unsigned long p;
		unsigned long size;
	
		//pos = READ_UINT32((unsigned long*)&Ztable[CD.sector * 6]);
		pos = (unsigned long)(Ztable[CD.sector * 6] + (Ztable[CD.sector * 6 + 1] << 8) + (Ztable[CD.sector * 6 + 2] << 16) + (Ztable[CD.sector * 6 + 3] << 24));
		fseek(CD.cd, pos, SEEK_SET);
		
		//p = READ_UINT16((unsigned short*)&Ztable[CD.sector * 6 + 4]);
		p = (unsigned short)(Ztable[CD.sector * 6 + 4] + (Ztable[CD.sector * 6 + 4 + 1] << 8));
		fread(Zbuf, 1, p, CD.cd);
		size = CD_FRAMESIZE_RAW;
			
		if( Z_OK != uncompress(CD.buffer, &size, Zbuf, p) )
		{
			SysPrintf("Error Uncompressing ROM");
		}
				
		CD.bufferPos = CD.sector;		
	}
	else if(Zmode == 2)
	{ // .ZNX
		unsigned long pos;
		unsigned long p;
		unsigned long size;
	
		//pos = READ_UINT32((unsigned long*)&Ztable[CD.sector * 6]);
		pos = (unsigned long)(Ztable[CD.sector * 10] + (Ztable[CD.sector * 10 + 1] << 8) + (Ztable[CD.sector * 10 + 2] << 16) + (Ztable[CD.sector * 10 + 3] << 24));
		fseek(CD.cd, pos, SEEK_SET);
		
		//p = READ_UINT16((unsigned short*)&Ztable[CD.sector * 6 + 4]);
		p = (unsigned short)(Ztable[CD.sector * 10 + 4] + (Ztable[CD.sector * 10 + 4 + 1] << 8));
		fread(Zbuf, 1, p, CD.cd);
		size = CD_FRAMESIZE_RAW;
			
		if( Z_OK != uncompress(CD.buffer, &size, Zbuf, p) )
		{
			SysPrintf("Error Uncompressing ROM");
		}
				
		CD.bufferPos = CD.sector;		
	}  
	else  	    
    {
        rc = fseek(CD.cd, CD.sector, SEEK_SET);
//        printf(" seek1 rc %d\n", rc);
        rc = fread(CD.buffer, 1, BUFFER_SIZE, CD.cd);
//        printf(" seek2 rc %d\n", rc);
    	CD.bufferPos = CD.sector;
    }

//    SysPrintf("end readit()\r\n");
}


void seekSector(const unsigned char m, const unsigned char s, const unsigned char f)
{
    // calc byte to search for
	if( Zmode )
	{		
		CD.sector = MSF2SECT(m, s, f); 
		//CD.sector = (( (m * 60) + (s - 2)) * 75 + f) * 2352;
    	// is it cached?
    	readit(m,s,f);
        return;
    }
    
    CD.sector = (( (m * 60) + (s - 2)) * 75 + f) * 2352;

    if ((CD.sector >= CD.bufferPos) &&
            (CD.sector < (CD.bufferPos + BUFFER_SIZE)) ) 
    {
        return;
    }
    // not cached - read a few blocks into the cache
    else
    {
        readit(m,s,f);
    }
}

long CDR_open(void)
{
	char str[256];
	
    //SysPrintf("start CDR_open()\r\n");

    //	newCD("/cd/cd.bin");
	strcpy(str, CDConfiguration.fn);
	
	CD.buffer = (unsigned char*)malloc(BUFFER_SIZE);
	
	if( CD.buffer == 0 )
	{
		SysPrintf("Not enough memory to buffer ROM");
	}
	
	UpdateZmode();
	
	newCD(str);

    //SysPrintf("end CDR_open()\r\n");
    return 0;
}

long CDR_init(void) {
	//char* ftrptr;

	strcpy(CDConfiguration.fn, packfile);
    return 0;
}

long CDR_shutdown(void) {
    return 0;
}

long CDR_close(void) {
    //SysPrintf("start CDR_close()\r\n");
    if( CD.cd )
    {
    	fclose(CD.cd);
    }
    if( CD.tl )
    {
    	free(CD.tl);
    	CD.tl = NULL;
    }
    
    if( CD.buffer )
	{
		free( CD.buffer );
		CD.buffer = NULL;
	}

	if (Ztable) 
	{ 
		free(Ztable); 
		Ztable = NULL;
		//FtrPtrFree(CREATORID, FTR_PSX_CACHE_ID);
	}
		
    //SysPrintf("end CDR_close()\r\n");
    return 0;
}

long CDR_getTN(unsigned char *buffer) {
//    SysPrintf("start CDR_getTN()\r\n");
//    SysPrintf("end CDR_getTN()\r\n");

	if( Zmode )
	{
		buffer[0] = 1;
		buffer[1] = 1;
		return 0;
	}
		
    return getTN(buffer);
//    return 0;
}

long CDR_getTD(unsigned char track, unsigned char *buffer) {
//    SysPrintf("start CDR_getTD()\r\n");
//    printf("getTD from track %d\n", track);

   unsigned char temp[3];
   int result;
   
   if( Zmode )
   {   
		buffer[2] = itob(0);
		buffer[1] = itob(2);
//		buffer[0] = 0;
		return 0;
   }
   
   result = getTD((int)track, temp);

   if (result == -1) return -1;

   buffer[1] = temp[1];
   buffer[2] = temp[0];

//    SysPrintf("end CDR_getTD()\r\n");
    return 0;
}

long CDR_readTrack(unsigned char *time) {
//    SysPrintf("start CDR_readTrack()\r\n");
//    printf("readTrack at %02d:%02d:%02d\n", BCDToInt(time[0]), BCDToInt(time[1]), BCDToInt(time[2]));

    //	cdrom_reinit();
	if( Zmode )
	{
		seekSector(btoi(time[0]), btoi(time[1]), btoi(time[2]));
	}
	else
	{
		seekSector(btoi(time[0]), btoi(time[1]), btoi(time[2]));
	}	
//    SysPrintf("end CDR_readTrack()\r\n");
    return 0;
}

unsigned char *CDR_getBuffer(void) {
//    SysPrintf("start CDR_getBuffer()\r\n");
    return getSector();
}

unsigned char* CALLBACK CDR_getBufferSub(void) { return NULL; }

long CALLBACK CDR_getStatus(struct CdrStat *stat) {
#if 0
    if (cdOpenCase) stat->Status = 0x10;
    else		stat->Status = 0;
#endif
	stat->Status = 0;

    return 0;
}

long CALLBACK CDR_play(unsigned char *sector) { return 0; }
long CALLBACK CDR_stop(void) { return 0; }

