/***********************************************************************
*
*	Dr.Hell's WinGDI GPU Plugin
*	Version 0.8
*	Copyright (C)Dr.Hell, 2002-2004
*
***********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include "../PsxCommon.h"
#include "../minimal.h"
#include <SDL/SDL.h>

//#define WITH_DRDEBUG
/*----------------------------------------------------------------------
Type
----------------------------------------------------------------------*/

typedef unsigned char uchar;
typedef char schar;
typedef unsigned short ushort;
typedef unsigned long ulong;


typedef struct {
	ulong Version;
	ulong GP1;
	ulong Control[256];
	uchar FrameBuffer[0x100000];
} GPUFREEZE;

/*----------------------------------------------------------------------
Constant
----------------------------------------------------------------------*/

#define	GPU_LIBTYPE			2
#define	GPU_VERSION			1
#define	GPU_REVISION		0
#define	GPU_BUILD			8

#define	FRAME_BUFFER_SIZE	(1024*512*2)
#define	FRAME_WIDTH			1024
#define	FRAME_HEIGHT		512

/*----------------------------------------------------------------------
Macro
----------------------------------------------------------------------*/

#define	FRAME_OFFSET(x,y)	(((y)<<10)+(x))

/*----------------------------------------------------------------------
Global
----------------------------------------------------------------------*/

int		px,py;
int		x_start,y_start,x_end,y_end;
ushort	*pvram;

char LIBNAME[] = "Dr.Hell's GDI Driver";
long gpuFreezeBegin;
long GP0;
long GP1;
long FrameToRead;
long FrameToWrite;
long FrameWidth;
long FrameCount;
long FrameIndex;
union {
	schar S1[64];
	short S2[32];
	long S4[16];
	uchar U1[64];
	ushort U2[32];
	ulong U4[16];
} PacketBuffer;
long PacketCount;
long PacketIndex;
long TextureWindow[4];
long DrawingArea[4];
long DrawingOffset[2];
long DisplayArea[8];
uchar Masking;
ushort PixelMSB;
long OtherEnv[16];
long gpuFreezeEnd;
long Skip = 0;
long DrawingCount[4];
long DisplayCount[4];
ushort *FrameBuffer;

/*----------------------------------------------------------------------
Table
----------------------------------------------------------------------*/

uchar PacketSize[256] = {
	0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	//		0-15
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	//		16-31
	3, 3, 3, 3, 6, 6, 6, 6, 4, 4, 4, 4, 8, 8, 8, 8,	//		32-47
	5, 5, 5, 5, 8, 8, 8, 8, 7, 7, 7, 7, 11, 11, 11, 11,	//	48-63
	2, 2, 2, 2, 0, 0, 0, 0, 3, 3, 3, 3, 3, 3, 3, 3,	//		64-79
	3, 3, 3, 3, 0, 0, 0, 0, 4, 4, 4, 4, 4, 4, 4, 4,	//		80-95
	2, 2, 2, 2, 3, 3, 3, 3, 1, 1, 1, 1, 2, 2, 2, 2,	//		96-111
	1, 1, 1, 1, 2, 2, 2, 2, 1, 1, 1, 1, 2, 2, 2, 2,	//		112-127
	3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	//		128-
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	//		144
	2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	//		160
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	//
	2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	//
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	//
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	//
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0	//
};

ushort HorizontalResolution[8] = {
	256, 368, 320, 384, 512, 512, 640, 640
};

ushort VerticalResolution[4] = {
	240, 480, 256, 480
};

/*----------------------------------------------------------------------
Device Depended
----------------------------------------------------------------------*/

#include "gpu_dev.c"

/*

CALLBACK
GetMem(p1)
FreeMem(p1)
CopyMem(p1,p2,p3)
ZeroMem(p1,p2)

long gpuGetTime(void);
void gpuLoadConfig(void);
void gpuSaveConfig(void);
void gpuOpenVideo(long);
void gpuCloseVideo(void);
void gpuVideoOutput(void);
void gpuOpenAbout(void);
void gpuOpenConfigure(void);

*/

/*----------------------------------------------------------------------
Drawing
----------------------------------------------------------------------*/

#include "gpu_draw.c"

/*

void gpuSendPacket(void);

*/

/*----------------------------------------------------------------------
Other
----------------------------------------------------------------------*/

/* gpuReset */
void gpuReset(void)
{
//	ZeroMem(&gpuFreezeBegin,
//			(long) &gpuFreezeEnd - (long) &gpuFreezeBegin);
	GP1 = 0x14802000;
	TextureWindow[2] = 255;
	TextureWindow[3] = 255;
	DrawingArea[2] = 256;
	DrawingArea[3] = 240;
	DisplayArea[2] = 256;
	DisplayArea[3] = 240;
	DisplayArea[6] = 256;
	DisplayArea[7] = 240;
}

/* ReadFrameBuffer */
/*
void ReadFrameBuffer(ushort * lpBuffer, long nCount)
{
	long temp, nCount2, nCount3;
	ushort *lpBuffer2;
	temp = FrameToRead;
	if (temp <= nCount) {
		nCount = temp;
		GP1 &= ~0x08000000;
	}
	FrameToRead = temp - nCount;
	lpBuffer2 = &FrameBuffer[FrameIndex];
	temp = FrameCount;
	while (nCount) {
		if (nCount >= temp) {
			nCount2 = temp;
			nCount3 = temp + 1024 - FrameWidth;
			temp = FrameWidth;
		} else {
			nCount3 = nCount2 = nCount;
			temp -= nCount;
		}
		CopyMem(lpBuffer, lpBuffer2, nCount2 + nCount2);
		lpBuffer += nCount2;
		lpBuffer2 += nCount3;
		nCount -= nCount2;
	}
	FrameCount = temp;
	FrameIndex = lpBuffer2 - FrameBuffer;
}
*/
/* WriteFrameBuffer */
/*
void WriteFrameBuffer(ushort * lpBuffer, long nCount)
{
	long temp, nCount2, nCount3;
	ushort *lpBuffer2;

	if (FrameToWrite <= nCount) {
		nCount = FrameToWrite;
		GP1 &= ~0x08000000;
	}
	FrameToWrite -= nCount;
	lpBuffer2 = &FrameBuffer[FrameIndex];

	while (nCount) {
		if (nCount >= FrameCount) {
			nCount2 = FrameCount;
			nCount3 = FrameCount + 1024 - FrameWidth;
			FrameCount = FrameWidth;
		} else {
			nCount3 = nCount2 = nCount;
			FrameCount -= nCount;
		}
		CopyMem(lpBuffer2, lpBuffer, nCount2 + nCount2);
		lpBuffer += nCount2;
		lpBuffer2 += nCount3;
		nCount -= nCount2;
	}

	FrameIndex = lpBuffer2 - FrameBuffer;
}
*/
/*----------------------------------------------------------------------
Plugin Functions
----------------------------------------------------------------------*/
long CALLBACK GPU_init(void);
long CALLBACK GPU_shutdown(void);
long CALLBACK GPU_open(long);
long CALLBACK GPU_close(void);
long CALLBACK GPU_configure(void);
void CALLBACK GPU_about(void);
long CALLBACK GPU_test(void);
void CALLBACK GPU_writeData(ulong);
void CALLBACK GPU_writeStatus(ulong);
ulong CALLBACK GPU_readData(void);
ulong CALLBACK GPU_readStatus(void);
long CALLBACK GPU_dmaChain(ulong *, ulong);
long CALLBACK GPU_getMode(void);
void CALLBACK GPU_setMode(ulong);
void CALLBACK GPU_updateLace(void);
void CALLBACK GPU_makeSnapshot(void);
void CALLBACK GPU_writeDataMem(ulong *, int);
void CALLBACK GPU_readDataMem(ulong *, int);
void CALLBACK GPU_displayText(char *);
void CALLBACK GPU_displayFlags(ulong);
long CALLBACK GPU_freeze(ulong, GPUFREEZE *);
void CALLBACK GPU_showScreenPic(uchar *);
void CALLBACK GPU_getScreenPic(uchar *);

/* GPUinit */
long CALLBACK GPU_init(void)
{
#ifdef WITH_DRDEBUG
dtext(10,10, "GPU_init START");
gp2x_video_flip();
#endif
	FrameBuffer = GetMem(FRAME_BUFFER_SIZE);
	gpuReset();
#ifdef WITH_DRDEBUG
dtext(10,10, "GPU_init END");
gp2x_video_flip();
#endif
	return (0);
}

/* GPUshutdown */
long CALLBACK GPU_shutdown(void)
{
	FreeMem(FrameBuffer);
	return (0);
}

/* GPUopen */
long CALLBACK GPU_open(long p1)
{
#ifdef WITH_DRDEBUG
dtext(10,10, "GPU_open START");
gp2x_video_flip();
#endif
	gpuOpenVideo(p1);
#ifdef WITH_DRDEBUG
dtext(10,10, "GPU_open END");
gp2x_video_flip();
#endif
	return (0);
}

/* GPUclose */
long CALLBACK GPU_close(void)
{
	gpuCloseVideo();
	return (0);
}

/* GPUconfigure */
long CALLBACK GPU_configure(void)
{
	return (gpuOpenConfigure());
}

/* GPUabout */
void CALLBACK GPU_about(void)
{
	gpuOpenAbout();
}

/* GPUtest */
long CALLBACK GPU_test(void)
{
	return (0);
}

/* GPUwriteData */
void CALLBACK GPU_writeData(ulong data)
{
#ifdef WITH_DRDEBUG
dtext(10,10, "WRITEDATA START");
gp2x_video_flip();
#endif
	GP1 &= ~0x14000000;

	if (FrameToWrite) {
          pvram[px]=(ushort)data;
          if (++px>=x_end) {
               px = x_start;
               pvram += 1024;
               if (++py>=y_end) FrameToWrite=0;
          }
          if (FrameToWrite) {
               pvram[px]=data>>16;
               if (++px>=x_end) {
                    px = x_start;
                    pvram += 1024;
                    if (++py>=y_end) FrameToWrite=0;
               }
          }
	} else {
		if (PacketCount) {
			PacketCount--;
			PacketBuffer.U4[PacketIndex++] = data;
		} else {
			PacketBuffer.U4[0] = data;
			PacketCount = PacketSize[data >> 24];
			PacketIndex = 1;
		}
		if (!PacketCount)
		{
			gpuSendPacket();
		}
	}
	GP1 |= 0x14000000;

#ifdef WITH_DRDEBUG
dtext(10,10, "WRITEDATA END");
gp2x_video_flip();
#endif
}

/* GPUwriteStatus */
void CALLBACK GPU_writeStatus(ulong data)
{
	switch (data >> 24) {
		case 0x00:
			gpuReset();
			break;
		case 0x01:
			GP1 &= ~0x08000000;
			PacketCount = FrameToRead = FrameToWrite = 0;
			break;
		case 0x02:
			GP1 &= ~0x08000000;
			PacketCount = FrameToRead = FrameToWrite = 0;
			break;
		case 0x03:
			GP1 = (GP1 & ~0x00800000) | ((data & 1) << 23);
			break;
		case 0x04:
			if (data == 0x04000000)
				PacketCount = 0;
			GP1 = (GP1 & ~0x60000000) | ((data & 3) << 29);
			break;
		case 0x05:
			DisplayArea[0] = data & 0x000003FF;
			DisplayArea[1] = (data & 0x000FFC00) >> 10;
			break;
		case 0x06:
			DisplayArea[4] = data & 0x00000FFF;
			DisplayArea[6] = (data & 0x00FFF000) >> 12;
			break;
		case 0x07:
			DisplayArea[5] = data & 0x000003FF;
			DisplayArea[7] = (data & 0x000FFC00) >> 10;
			break;
		case 0x08:
			OtherEnv[0x08] = (data >> 7) & 1;	//	reverse(?)
			GP1 =
				(GP1 & ~0x007F0000) | ((data & 0x3F) << 17) |
				((data & 0x40) << 10);
			DisplayArea[2] = HorizontalResolution[(GP1 >> 16) & 7];
			DisplayArea[3] = VerticalResolution[(GP1 >> 19) & 3];
			break;
		case 0x09:
			OtherEnv[0x09] = data & 1;			//	gpub(?)
			break;
		case 0x10:
			switch (data & 0xffff) {
				case 0:
				case 1:
				case 3:
					GP0 = (DrawingArea[1] << 10) | DrawingArea[0];
					break;
				case 4:
					GP0 =
						((DrawingArea[3] - 1) << 10) | (DrawingArea[2] -
														1);
					break;
				case 6:
				case 5:
					GP0 = (DrawingOffset[1] << 11) | DrawingOffset[0];
					break;
				case 7:
					GP0 = 2;
					break;
				default:
					GP0 = 0;
			}
			break;
	}
}

/* GPUreadData */
ulong CALLBACK GPU_readData(void)
{
#ifdef WITH_DRDEBUG
dtext(10,10, "READDATA START");
gp2x_video_flip();
#endif
	GP1 &= ~0x14000000;
	if (FrameToRead)
	{
		GP0 = pvram[px];
		if (++px>=x_end) {
		   px = x_start;
		   pvram += 1024;
		   if (++py>=y_end) FrameToRead=0;
		}
		GP0 |= pvram[px]<<16;
		if (++px>=x_end) {
		   px = x_start;
		   pvram +=1024;
		   if (++py>=y_end) FrameToRead=0;
		}

		if( FrameToRead == 0 ) GP1 &= ~0x08000000;
	}
	GP1 |= 0x14000000;
#ifdef WITH_DRDEBUG
dtext(10,10, "READDATA END");
gp2x_video_flip();
#endif
	return (GP0);
}

/* GPUreadStatus */
ulong CALLBACK GPU_readStatus(void)
{
	return (GP1);
}

/* GPUdmaChain */
long CALLBACK GPU_dmaChain(ulong * baseAddr, ulong dmaVAddr)
{
	ulong temp, data, pkts, *address, count, offset;
#ifdef WITH_DRDEBUG
dtext(10,10, "GPU_dmaChain START");
gp2x_video_flip();
#endif
	GP1 &= ~0x14000000;
	dmaVAddr &= 0x00FFFFFF;
	pkts = 0;
	while (dmaVAddr != 0xFFFFFF) {
		address = (baseAddr + (dmaVAddr >> 2));
		data = *address++;
		count = (data >> 24);
		offset = data & 0x00FFFFFF;
		if (dmaVAddr != offset)
			dmaVAddr = offset;
		else
			dmaVAddr = 0xFFFFFF;
		while (count) {
			data = *address++;
			count--;
			//temp = PacketCount;
			if (PacketCount) {
				PacketCount--;
				PacketBuffer.U4[PacketIndex++] = data;
			} else {
				PacketBuffer.U4[0] = data;
				PacketCount = PacketSize[data >> 24];
				PacketIndex = 1;
			}
			//PacketCount = temp;
			if (!PacketCount)
			{
				gpuSendPacket();
			}
		}
		pkts++;
		if (pkts > 0x100000) return(dmaVAddr);
	}
	GP1 = (GP1 | 0x14000000) & ~0x60000000;
#ifdef WITH_DRDEBUG
dtext(10,10, "GPU_dmaChain END");
gp2x_video_flip();
#endif
	return (0);
}

/* GPUgetMode */
long CALLBACK GPU_getMode(void)
{
	/* Support Discontinued */
	return (0);
}

/* GPUsetMode */
void CALLBACK GPU_setMode(ulong p1)
{
	/* Support Discontinued */
}

/* GPUupdateLace */
void CALLBACK GPU_updateLace(void)
{
	long temp;

#ifdef WITH_DRDEBUG
dtext(10,10, "GPU_updateLace START");
gp2x_video_flip();
#endif

	GP1 ^= 0x80000000;

#if 0
	DisplayCount[1] += DisplayCount[0];
	DrawingCount[2] &= 0xFF;
	DrawingCount[3] &= 0xFF;
	Skip = DrawingCount[2];
	if (DrawingCount[2]) {
		DrawingCount[2]--;
		Skip = DrawingCount[2];
		return;
	}
	if (DrawingCount[0]) {
		if (DisplayArea[3] > 240) {
			if (DrawingCount[3]) {
				DrawingCount[3]--;
				return;
			}
			DrawingCount[3] = DrawingCount[1];
		}
	}
	DrawingCount[2] = DrawingCount[0];
#endif

	gpuVideoOutput();
	
#if 0
	temp = gpuGetTime() * 1000;
	if ((ulong) (temp - DisplayCount[2]) > (ulong) 10000000) {
		DisplayCount[1] = 0;
		DisplayCount[2] = temp;
	} else {
		while ((temp - DisplayCount[2]) < DisplayCount[1]) {
			temp = gpuGetTime() * 1000;
		}
		if (DisplayCount[1] > 5000000) {
			DisplayCount[1] = 0;
			DisplayCount[2] = temp;
		}
	}
#endif

#ifdef WITH_DRDEBUG
dtext(10,10, "GPU_updateLace END");
gp2x_video_flip();
#endif
}

/* GPUmakeSnapshot */
void CALLBACK GPU_makeSnapshot(void)
{
	/* Support Discontinued */
}

/* GPUwriteDataMem */
void CALLBACK GPU_writeDataMem(ulong * dmaAddress, int dmaCount)
{
	ulong temp, temp2;
	GP1 &= ~0x14000000;
	while (dmaCount) {
		if (FrameToWrite) {
            while (dmaCount--) 
			{
                ulong data = *dmaAddress++;

                if (px<1024 && py<512)
                    pvram[px] = data;
                if (++px>=x_end) 
				{
                    px = x_start;
                    pvram += 1024;
                    if (++py>=y_end) 
					{
                        FrameToWrite = 0;
						GP1 &= ~0x08000000;
                        break;
                    }
                }
                if (px<1024 && py<512)
                    pvram[px] = data>>16;
                if (++px>=x_end) 
				{
                    px = x_start;
                    pvram += 1024;
                    if (++py>=y_end) 
					{
                        FrameToWrite = 0;
						GP1 &= ~0x08000000;
                        break;
                    }
                }
            }
		} else {
			temp = *dmaAddress++;
			dmaCount--;
			if (PacketCount) {
				PacketBuffer.U4[PacketIndex++] = temp;
				PacketCount--;
			} else {
				PacketBuffer.U4[0] = temp;
				PacketCount = PacketSize[temp >> 24];
				PacketIndex = 1;
			}
			if (!PacketCount)
			{
				gpuSendPacket();
			}
		}
	}
	GP1 = (GP1 | 0x14000000) & ~0x60000000;
}

/* GPUreadDataMem */
void CALLBACK GPU_readDataMem(ulong * dmaAddress, int dmaCount)
{
	if( FrameToRead == 0 ) return;

	GP1 &= ~0x14000000;

	do 
	{
		// lower 16 bit
		ulong data = (unsigned long)pvram[px];

		if (++px>=x_end) 
		{
			px = x_start;
			pvram += 1024;
		}

		// higher 16 bit (always, even if it's an odd width)
		data |= (unsigned long)(pvram[px])<<16;
    
		*dmaAddress++ = data;

		if (++px>=x_end) 
		{
			px = x_start;
			pvram += 1024;
			if (++py>=y_end) 
			{
				FrameToRead = 0;
				GP1 &= ~0x08000000;
				break;
			}
		}
	} while (--dmaCount);

#ifdef WITH_DRDEBUG
dtext(10,10, "READDATAMEM END");
gp2x_video_flip();
#endif
	GP1 = (GP1 | 0x14000000) & ~0x60000000;
}

/* GPUdisplayText */
void CALLBACK GPU_displayText(char *p1)
{
	/* Support Discontinued */
}

/* GPUdisplayFlags */
void CALLBACK GPU_displayFlags(ulong p1)
{
	/* Support Discontinued */
}

/* GPUfreeze */
long CALLBACK GPU_freeze(ulong p1, GPUFREEZE * p2)
{
	ulong temp;
	if (p1 == 2) {
		temp = *(ulong *) p2;
		if ((temp < 0) || (temp > 8))
			return (0);
		return (1);
	}
	if (p2 == NULL)
		return (0);
	if (p2->Version != 1)
		return (0);
	if (p1 == 1) {
		p2->GP1 = GP1;
		//CopyMem(p2->Control, &gpuFreezeBegin,
		//		(long) &gpuFreezeEnd - (long) &gpuFreezeBegin);
		//CopyMem(p2->FrameBuffer, FrameBuffer, FRAME_BUFFER_SIZE);
		return (1);
	}
	if (p1 == 0) {
		GP1 = p2->GP1;
		//CopyMem(&gpuFreezeBegin, p2->Control,
		//		(long) &gpuFreezeEnd - (long) &gpuFreezeBegin);
		//CopyMem(FrameBuffer, p2->FrameBuffer, FRAME_BUFFER_SIZE);
		return (1);
	}
	return (0);
}

/* GPUshowScreenPic */
void CALLBACK GPU_showScreenPic(uchar * p1)
{
	/* Support Discontinued */
}

/* GPUgetScreenPic */
void CALLBACK GPU_getScreenPic(uchar * p1)
{
	/* Support Discontinued */
}

