/*  Pcsx - Pc Psx Emulator
 *  Copyright (C) 1999-2003  Pcsx Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __PLUGINS_H__
#define __PLUGINS_H__

typedef void* HWND;
#define CALLBACK
//#include "plugin.h"


#include "PSEmu_Plugin_Defs.h"
#include "Decode_XA.h"

int  LoadPlugins();
void ReleasePlugins();
int  OpenPlugins();
void ClosePlugins();
void ResetPlugins();


unsigned long PSEgetLibType(void);
unsigned long PSEgetLibVersion(void);
char* PSEgetLibName(void);

///GPU PLUGIN STUFF 
long GPU_open(long);
long GPU_init(void);
long GPU_shutdown(void);
long GPU_close(void);
void GPU_writeStatus(unsigned long);
void GPU_writeData(unsigned long);
void GPU_writeDataMem(unsigned long *, int);
unsigned long GPU_readStatus(void);
unsigned long GPU_readData(void);
void GPU_readDataMem(unsigned long *, int);
long GPU_dmaChain(unsigned long *,unsigned long);
void GPU_updateLace(void);
long GPU_configure(void);
long GPU_test(void);
void GPU_about(void);
void GPU_makeSnapshot(void);
void GPU_keypressed(int);
void GPU_displayText(char *);

typedef struct {
	unsigned long ulFreezeVersion;
	unsigned long ulStatus;
	unsigned long ulControl[256];
	unsigned char psxVRam[1024*512*2];
} GPUFreeze_t;

void GPU_getScreenPic(unsigned char *);
void GPU_showScreenPic(unsigned char *);

//cd rom plugin ;)
long CDR_init(void);
long CDR_shutdown(void);
long CDR_open(void);
long CDR_close(void);
long CDR_getTN(unsigned char *);
long CDR_getTD(unsigned char , unsigned char *);
long CDR_readTrack(unsigned char *);
unsigned char* CDR_getBuffer(void);
long CDR_configure(void);
long CDR_test(void);
void CDR_about(void);
long CDR_play(unsigned char *);
long CDR_stop(void);
struct CdrStat {
	unsigned long Type;
	unsigned long Status;
	unsigned char Time[3];
};
long CDR_getStatus(struct CdrStat *);
char* CDR_getDriveLetter(void);
struct SubQ {
	char res0[11];
	unsigned char ControlAndADR;
	unsigned char TrackNumber;
	unsigned char IndexNumber;
	unsigned char TrackRelativeAddress[3];
	unsigned char Filler;
	unsigned char AbsoluteAddress[3];
	char res1[72];
};
unsigned char* CDR_getBufferSub(void);

// SPU_ plugin
long SPU_open(void);
long SPU_init(void);				
long SPU_shutdown(void);	
long SPU_close(void);			
void SPU_playSample(unsigned char);		
void SPU_startChannels1(unsigned short);	
void SPU_startChannels2(unsigned short);
void SPU_stopChannels1(unsigned short);	
void SPU_stopChannels2(unsigned short);	
void SPU_putOne(unsigned long,unsigned short);			
unsigned short SPU_getOne(unsigned long);			
void SPU_setAddr(unsigned char, unsigned short);			
void SPU_setPitch(unsigned char, unsigned short);		
void SPU_setVolumeL(unsigned char, short );		
void SPU_setVolumeR(unsigned char, short );		
//psemu pro 2 functions from now..
void SPU_writeRegister(unsigned long, unsigned short);	
unsigned short SPU_readRegister(unsigned long);		
void SPU_writeDMA(unsigned short);
unsigned short SPU_readDMA(void);
void SPU_writeDMAMem(unsigned short *, int);
void SPU_readDMAMem(unsigned short *, int);
void SPU_playADPCMchannel(xa_decode_t *);
void SPU_registerCallback(void);
long SPU_configure(void);
long SPU_test(void);			
void SPU_about(void);
typedef struct {
	unsigned char PluginName[8];
	unsigned long PluginVersion;
	unsigned long Size;
	unsigned char SPUPorts[0x200];
	unsigned char SPURam[0x80000];
	xa_decode_t xa;
	unsigned char *SPUInfo;
} SPUFreeze_t;
long SPU_freeze(unsigned long, SPUFreeze_t *);
void SPU_async(unsigned long);

// PAD Functions
long PAD1_open(void);
long PAD1_configure(void);
void PAD1_about(void);
long PAD1_init(long);
long PAD1_shutdown(void);
long PAD1_test(void);		
long PAD1_close(void);
long PAD1_query(void);
long PAD1_readPort1(PadDataS*);
long PAD1_readPort2(PadDataS*);
long PAD1_keypressed(void);
unsigned char PAD1_startPoll(int);
unsigned char PAD1_poll(unsigned char);
void PAD1_setSensitive(int);

long PAD2_open(void);
long PAD2_configure(void);
void PAD2_about(void);
long PAD2_init(long);
long PAD2_shutdown(void);
long PAD2_test(void);		
long PAD2_close(void);
long PAD2_query(void);
long PAD2_readPort1(PadDataS*);
long PAD2_readPort2(PadDataS*);
long PAD2_keypressed(void);
unsigned char PAD2_startPoll(int);
unsigned char PAD2_poll(unsigned char);
void PAD2_setSensitive(int);

int LoadCDRplugin(char *CDRdll);
int LoadGPUplugin(char *GPUdll);
int LoadSPUplugin(char *SPUdll);
int LoadPAD1plugin(char *PAD1dll);
int LoadPAD2plugin(char *PAD2dll);
int LoadNETplugin(char *NETdll);

void CALLBACK clearDynarec(void);

#endif /* __PLUGINS_H__ */
